from __future__ import annotations

import sys
import typing as t

if t.TYPE_CHECKING:
    from arc.internal.types import ClientT


@t.overload
def loader() -> t.Callable[[t.Callable[[ClientT], None]], t.Callable[[ClientT], None]]: ...


@t.overload
def loader(callback: t.Callable[[ClientT], None]) -> t.Callable[[ClientT], None]: ...


def loader(
    callback: t.Callable[[ClientT], None] | None = None,
) -> t.Callable[[ClientT], None] | t.Callable[[t.Callable[[ClientT], None]], t.Callable[[ClientT], None]]:
    """Decorator to set the load callback for this module.

    Example
    --------
    ```py
    client.load_extension("my_extension")

    # In my_extension.py:

    @arc.loader
    def load(client: arc.GatewayClient) -> None:
        client.add_plugin(...)
    ```

    See Also
    --------
    - [`Client.load_extension`][arc.client.Client.load_extension]
    """

    def decorator(func: t.Callable[[ClientT], None]) -> t.Callable[[ClientT], None]:
        module = sys.modules[func.__module__]
        setattr(module, "__arc_extension_loader__", func)
        return func

    if callback is not None:
        return decorator(callback)

    return decorator


@t.overload
def unloader() -> t.Callable[[t.Callable[[ClientT], None]], t.Callable[[ClientT], None]]: ...


@t.overload
def unloader(callback: t.Callable[[ClientT], None]) -> t.Callable[[ClientT], None]: ...


def unloader(
    callback: t.Callable[[ClientT], None] | None = None,
) -> t.Callable[[ClientT], None] | t.Callable[[t.Callable[[ClientT], None]], t.Callable[[ClientT], None]]:
    """Decorator to set the unload callback for this module.

    Example
    --------
    ```py
    client.unload_extension("my_extension")

    # In my_extension.py:

    @arc.unloader
    def unload(client: arc.GatewayClient) -> None:
        client.remove_plugin(...)
    ```

    See Also
    --------
    - [`Client.unload_extension`][arc.client.Client.unload_extension]
    """

    def decorator(func: t.Callable[[ClientT], None]) -> t.Callable[[ClientT], None]:
        module = sys.modules[func.__module__]
        setattr(module, "__arc_extension_unloader__", func)
        return func

    if callback is not None:
        return decorator(callback)

    return decorator


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import asyncio
import sys
import time
import traceback
import typing as t
from collections import deque

import attr

from arc.abc.hookable import HookResult

__all__ = ("RateLimiter", "RateLimiterExhaustedError")

KeyT = t.TypeVar("KeyT")


@attr.define(slots=True, kw_only=True)
class _Bucket(t.Generic[KeyT]):
    """Handles the ratelimiting of a single item. (E.g. a single user or a channel)."""

    key: str
    """The key of the bucket."""

    reset_at: float
    """The time at which the bucket resets."""

    limiter: RateLimiter[KeyT]
    """The limiter this bucket belongs to."""

    _remaining: int = attr.field(alias="remaining")
    """The amount of requests remaining until the bucket is exhausted."""

    _queue: deque[asyncio.Event] = attr.field(factory=deque, init=False)
    """A list of events to set as the iter task proceeds."""

    _task: asyncio.Task[None] | None = attr.field(default=None, init=False)
    """The task that is currently iterating over the queue."""

    @classmethod
    def for_limiter(cls, key: str, limiter: RateLimiter[KeyT]) -> _Bucket[KeyT]:
        """Create a new bucket for a RateLimiter."""
        return cls(key=key, limiter=limiter, reset_at=time.monotonic() + limiter.period, remaining=limiter.limit)

    @property
    def remaining(self) -> int:
        """The amount of requests remaining until the bucket is exhausted."""
        if self.reset_at <= time.monotonic():
            self.reset()
        return self._remaining

    @remaining.setter
    def remaining(self, value: int) -> None:
        self._remaining = value

    @property
    def is_exhausted(self) -> bool:
        """Return a boolean determining if the bucket is exhausted."""
        return self.remaining <= 0 and self.reset_at > time.monotonic()

    @property
    def is_stale(self) -> bool:
        """Return a boolean determining if the bucket is stale.
        If a bucket is stale, it is no longer in use and can be purged.
        """
        return not self._queue and self.remaining == self.limiter.limit and (self._task is None or self._task.done())

    def start_queue(self) -> None:
        """Start the queue of a bucket.
        This will start setting events in the queue until the bucket is ratelimited.
        """
        if self._task is None or self._task.done():
            self._task = asyncio.create_task(self._iter_queue())

    def reset(self) -> None:
        """Reset the bucket."""
        self.reset_at = time.monotonic() + self.limiter.period
        self._remaining = self.limiter.limit

    async def _iter_queue(self) -> None:
        """Iterate over the queue and set events until exhausted."""
        try:
            while self._queue:
                if self.remaining <= 0 and self.reset_at > time.monotonic():
                    # Sleep until ratelimit expires
                    await asyncio.sleep(self.reset_at - time.monotonic())
                    self.reset()

                # Set events while not ratelimited
                while self.remaining > 0 and self._queue:
                    self._queue.popleft().set()
                    self._remaining -= 1

        except Exception as e:
            print(f"Task Exception was never retrieved: {e}", file=sys.stderr)
            print(traceback.format_exc(), file=sys.stderr)


class RateLimiter(t.Generic[KeyT]):
    """A fixed window ratelimiter implementation.

    It is generic over the type of object it extracts the key from (`KeyT`), and must also be given a key extraction function.
    It can be used independently for any purpose, or can be used as a hook via [`LimiterHook`][arc.utils.hooks.limiters.LimiterHook].

    !!! tip
        If you're looking to use this ratelimiter implementation as a hook, see [`LimiterHook`][arc.utils.hooks.limiters.LimiterHook]
        along with all the other built-in limiter hooks.

    Parameters
    ----------
    period : float
        The period, in seconds, after which the bucket resets.
    limit : int
        The amount of requests allowed in a bucket.
    get_key_with : t.Callable[[Context[t.Any]], str]
        A callable that returns a key for the ratelimiter bucket.
    """

    __slots__ = ("_buckets", "_gc_task", "_get_key", "limit", "period")

    def __init__(self, period: float, limit: int, *, get_key_with: t.Callable[[KeyT], str]) -> None:
        self.period: float = period
        self.limit: int = limit
        self._buckets: dict[str, _Bucket[KeyT]] = {}
        self._get_key: t.Callable[[KeyT], str] = get_key_with
        self._gc_task: asyncio.Task[None] | None = None

    def get_key(self, item: KeyT) -> str:
        """Get key for ratelimiter bucket."""
        return self._get_key(item)

    def is_rate_limited(self, item: KeyT) -> bool:
        """Returns a boolean determining if the ratelimiter is ratelimited or not.

        Parameters
        ----------
        item : Context[t.Any]
            The context to evaluate the ratelimit under.

        Returns
        -------
        bool
            A boolean determining if the ratelimiter is ratelimited or not.
        """
        now = time.monotonic()

        if data := self._buckets.get(self.get_key(item)):
            if data.reset_at <= now:
                return False
            return data._remaining <= 0
        return False

    def _start_gc(self) -> None:
        """Start the garbage collector task if one is not running."""
        if self._gc_task is None or self._gc_task.done():
            self._gc_task = asyncio.create_task(self._gc())

    async def _gc(self) -> None:
        """Purge stale buckets."""
        while self._buckets:
            await asyncio.sleep(self.period + 1.0)
            for bucket in list(self._buckets.values()):
                if bucket.is_stale:
                    del self._buckets[bucket.key]

    async def acquire(self, item: KeyT, /, *, wait: bool = True) -> None:
        """Acquire a bucket, block execution if ratelimited and wait is True.

        Parameters
        ----------
        item : KeyT
            The item to evaluate the ratelimit under.
        wait : bool
            Determines if this call should block in
            case of hitting a ratelimit.

        Raises
        ------
        RateLimiterExhaustedError
            If the ratelimiter is ratelimited and wait is False.
        """
        event = asyncio.Event()

        key = self.get_key(item)
        # Get existing or insert new bucket
        bucket = self._buckets.setdefault(key, _Bucket.for_limiter(key, self))  # pyright: ignore reportUnknowMemberType

        if bucket.is_exhausted and not wait:
            raise RateLimiterExhaustedError(
                self,
                bucket.reset_at - time.monotonic(),
                f"Ratelimited for {bucket.reset_at - time.monotonic()} seconds.",
            )

        bucket._queue.append(event)
        bucket.start_queue()
        self._start_gc()

        if wait:
            await event.wait()

    async def __call__(self, item: KeyT) -> HookResult:
        """Acquire a ratelimit, fail if ratelimited.

        Parameters
        ----------
        item : KeyT
            The context to evaluate the ratelimit under.

        Returns
        -------
        HookResult
            A hook result to conform to the limiter protocol.

        Raises
        ------
        UnderCooldownError
            If the ratelimiter is ratelimited.
        """
        await self.acquire(item, wait=False)
        return HookResult()

    def reset(self, ctx: KeyT) -> None:
        """Reset the ratelimit for a given context."""
        if bucket := self._buckets.get(self.get_key(ctx)):
            bucket.reset()


class RateLimiterExhaustedError(Exception):
    """Raised when a [RateLimiter][arc.utils.ratelimiter.RateLimiter] is acquired while it is exhausted, and the
    `wait` parameter is set to `False`.

    Attributes
    ----------
    retry_after : float
        The amount of time in seconds until the command is off cooldown.
    limiter : arc.abc.limiter.LimiterProto
        The limiter that was rate limited.
    """

    def __init__(self, limiter: RateLimiter[t.Any], retry_after: float, *args: t.Any) -> None:
        self.retry_after = retry_after
        self.limiter = limiter
        super().__init__(*args)


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

# arc uses dependency injection to manage state across commands
# Related documentation guide: https://arc.hypergonial.com/guides/dependency_injection


# This is just an example "database" that stores a single integer
class MyDatabase:
    def __init__(self, value: int) -> None:
        self.value = value


bot = hikari.GatewayBot("...")
client = arc.GatewayClient(bot)

# Create a new instance of 'MyDatabase'
database = MyDatabase(value=0)

# We declare a new dependency of type 'MyDatabase' and the value of 'database'
client.set_type_dependency(MyDatabase, database)


@client.include
@arc.slash_command("increment", "Increment a counter!")
# We inject a dependency of type 'MyDatabase' here.
async def increment(ctx: arc.GatewayContext, db: MyDatabase = arc.inject()) -> None:
    db.value += 1
    await ctx.respond(f"Counter is at: `{db.value}`")


# You can also inject dependencies into functions that aren't commands
@client.inject_dependencies
def compare_counter(value: int, db: MyDatabase = arc.inject()) -> None:
    if value > db.value:
        print("Value is bigger!")
    else:
        print("Counter is bigger or equal!")


# Note that using 'arc.inject()' outside a command, or a function decorated with 'arc.inject_dependencies'
# will lead to unexpected behavior.

# Commands are automatically injected with dependencies, thus using '@arc.inject_dependencies' on them is redundant.

bot.run()



from __future__ import annotations

import typing as t
from functools import total_ordering

from arc.internal.about import __version__

if t.TYPE_CHECKING:
    import typing_extensions as te


__all__ = ("CURRENT_VERSION", "Version")


@total_ordering
class Version:
    """A version object representing a semantic version."""

    __slots__: t.Sequence[str] = ("major", "minor", "patch")

    def __init__(self, major: int, minor: int, patch: int) -> None:
        self.major = major
        self.minor = minor
        self.patch = patch

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def __lt__(self, other: te.Self) -> bool:
        return (self.major, self.minor, self.patch) < (other.major, other.minor, other.patch)

    @classmethod
    def from_str(cls, version: str) -> te.Self:
        if version[0].casefold() == "v":
            version = version[1:]

        major, minor, patch = version.split(".", maxsplit=2)
        return cls(int(major), int(minor), int(patch))


CURRENT_VERSION: Version = Version.from_str(__version__)

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



"""A command handler for hikari.

To get started, see the following links:

GitHub:
https://github.com/hypergonial/hikari-arc
Documentation:
https://arc.hypergonial.com
"""

from alluka import Client as Injector
from alluka import OverridingContext as InjectorOverridingContext
from alluka import inject

from arc import abc, command, ext, utils

from .abc import HookResult, Option, OptionType, with_concurrency_limit, with_hook, with_post_hook
from .client import (
    GatewayClient,
    GatewayClientBase,
    GatewayContext,
    GatewayPlugin,
    RESTClient,
    RESTClientBase,
    RESTContext,
    RESTPlugin,
)
from .command import (
    AttachmentParams,
    BoolParams,
    ChannelParams,
    ColorParams,
    ColourParams,
    EmojiParams,
    FloatParams,
    IntParams,
    MemberParams,
    MentionableParams,
    MessageCommand,
    RoleParams,
    SlashCommand,
    SlashGroup,
    SlashSubCommand,
    SlashSubGroup,
    StrParams,
    UserCommand,
    UserParams,
    message_command,
    slash_command,
    slash_subcommand,
    user_command,
)
from .context import AutocompleteData, AutodeferMode, Context, InteractionResponse
from .errors import (
    ArcError,
    AutocompleteError,
    BotMissingPermissionsError,
    CommandInvokeError,
    CommandPublishFailedError,
    DMOnlyError,
    ExtensionError,
    ExtensionLoadError,
    ExtensionUnloadError,
    GlobalCommandPublishFailedError,
    GuildCommandPublishFailedError,
    GuildOnlyError,
    HookAbortError,
    InteractionResponseError,
    InvokerMissingPermissionsError,
    MaxConcurrencyReachedError,
    NoResponseIssuedError,
    NotOwnerError,
    OptionConverterFailureError,
    ResponseAlreadyIssuedError,
    UnderCooldownError,
)
from .events import ArcEvent, CommandErrorEvent, StartedEvent, StoppingEvent
from .extension import loader, unloader
from .internal.about import __version__
from .locale import (
    CommandLocaleRequest,
    CustomLocaleRequest,
    LocaleRequest,
    LocaleRequestType,
    LocaleResponse,
    OptionLocaleRequest,
)
from .plugin import GatewayPluginBase, PluginBase, RESTPluginBase
from .utils import (
    bot_has_permissions,
    channel_concurrency,
    channel_limiter,
    custom_concurrency,
    custom_limiter,
    dm_only,
    global_concurrency,
    global_limiter,
    guild_concurrency,
    guild_limiter,
    guild_only,
    has_permissions,
    member_concurrency,
    member_limiter,
    owner_only,
    user_concurrency,
    user_limiter,
)

__all__ = (
    "ArcError",
    "ArcEvent",
    "AttachmentParams",
    "AutocompleteData",
    "AutocompleteError",
    "AutodeferMode",
    "BoolParams",
    "BotMissingPermissionsError",
    "ChannelParams",
    "ColorParams",
    "ColourParams",
    "CommandErrorEvent",
    "CommandInvokeError",
    "CommandLocaleRequest",
    "CommandPublishFailedError",
    "Context",
    "CustomLocaleRequest",
    "DMOnlyError",
    "EmojiParams",
    "ExtensionError",
    "ExtensionLoadError",
    "ExtensionUnloadError",
    "FloatParams",
    "GatewayClient",
    "GatewayClientBase",
    "GatewayContext",
    "GatewayPlugin",
    "GatewayPluginBase",
    "GlobalCommandPublishFailedError",
    "GuildCommandPublishFailedError",
    "GuildOnlyError",
    "HookAbortError",
    "HookResult",
    "Injector",
    "InjectorOverridingContext",
    "IntParams",
    "InteractionResponse",
    "InteractionResponseError",
    "InvokerMissingPermissionsError",
    "LocaleRequest",
    "LocaleRequestType",
    "LocaleResponse",
    "MaxConcurrencyReachedError",
    "MemberParams",
    "MentionableParams",
    "MessageCommand",
    "NoResponseIssuedError",
    "NotOwnerError",
    "Option",
    "OptionConverterFailureError",
    "OptionLocaleRequest",
    "OptionType",
    "PluginBase",
    "RESTClient",
    "RESTClientBase",
    "RESTContext",
    "RESTPlugin",
    "RESTPluginBase",
    "ResponseAlreadyIssuedError",
    "RoleParams",
    "SlashCommand",
    "SlashGroup",
    "SlashSubCommand",
    "SlashSubGroup",
    "StartedEvent",
    "StoppingEvent",
    "StrParams",
    "UnderCooldownError",
    "UserCommand",
    "UserParams",
    "__version__",
    "abc",
    "bot_has_permissions",
    "channel_concurrency",
    "channel_limiter",
    "command",
    "custom_concurrency",
    "custom_limiter",
    "dm_only",
    "ext",
    "global_concurrency",
    "global_limiter",
    "guild_concurrency",
    "guild_limiter",
    "guild_only",
    "has_permissions",
    "inject",
    "loader",
    "member_concurrency",
    "member_limiter",
    "message_command",
    "owner_only",
    "slash_command",
    "slash_subcommand",
    "unloader",
    "user_command",
    "user_concurrency",
    "user_limiter",
    "utils",
    "with_concurrency_limit",
    "with_hook",
    "with_post_hook",
)

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

# Extensions are a way to split your code into multiple files
# Related documentation guide: https://arc.hypergonial.com/guides/plugin_extensions

bot = hikari.RESTBot("...")
client = arc.RESTClient(bot)

# Load all extensions located in the 'extensions' directory
client.load_extensions_from("extensions")
# You can also use 'client.load_extension' to load a single extension.

bot.run()



from __future__ import annotations

import abc
import typing as t

from arc.internal.types import ClientT

if t.TYPE_CHECKING:
    from arc.abc.hookable import HookResult
    from arc.context.base import Context


@t.runtime_checkable
class LimiterProto(t.Protocol, t.Generic[ClientT]):
    """A protocol that all limiter hooks should implement.
    A limiter is simply a special type of hook with added methods.

    If you're looking to integrate your own ratelimiter implementation,
    you should make sure to implement all methods defined here.

    !!! tip
        An easy (but not necessary) way to ensure you've implemented all methods
        is to inherit from this protocol.
    """

    @abc.abstractmethod
    async def __call__(self, ctx: Context[ClientT], /) -> HookResult:
        """Call the limiter with the given context.
        Implementations should raise an exception if the limiter is ratelimited
        or abort via [`HookResult`][arc.abc.hookable.HookResult].

        Parameters
        ----------
        ctx : Context
            The context to evaluate the ratelimit under.
        """

    @abc.abstractmethod
    def reset(self, ctx: Context[ClientT], /) -> None:
        """Reset the limiter for the given context.

        Parameters
        ----------
        ctx : Context
            The context to reset the ratelimit for.
        """

    @abc.abstractmethod
    def is_rate_limited(self, ctx: Context[ClientT], /) -> bool:
        """Check if the limiter is rate limited for the given context.

        Parameters
        ----------
        ctx : Context
            The context to evaluate the ratelimit under.

        Returns
        -------
        bool
            Whether or not the limiter is ratelimited.
        """



from .attachment import AttachmentOption, AttachmentParams
from .bool import BoolOption, BoolParams
from .channel import ChannelOption, ChannelParams
from .custom import (
    ColorOption,
    ColorParams,
    ColourOption,
    ColourParams,
    EmojiOption,
    EmojiParams,
    MemberOption,
    MemberParams,
)
from .float import FloatOption, FloatParams
from .int import IntOption, IntParams
from .mentionable import MentionableOption, MentionableParams
from .role import RoleOption, RoleParams
from .str import StrOption, StrParams
from .user import UserOption, UserParams

__all__ = (
    "AttachmentOption",
    "AttachmentParams",
    "BoolOption",
    "BoolParams",
    "ChannelOption",
    "ChannelParams",
    "ColorOption",
    "ColorParams",
    "ColourOption",
    "ColourParams",
    "EmojiOption",
    "EmojiParams",
    "FloatOption",
    "FloatParams",
    "IntOption",
    "IntParams",
    "MemberOption",
    "MemberParams",
    "MentionableOption",
    "MentionableParams",
    "RoleOption",
    "RoleParams",
    "StrOption",
    "StrParams",
    "UserOption",
    "UserParams",
)

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

# Limiters are a special type of pre-execution hook that can be used to limit
# the amount of times a command can be executed in a given period. You may also know them as cooldowns.

# Related documentation guide: https://arc.hypergonial.com/guides/hooks#limiters

bot = hikari.RESTBot("...")
client = arc.RESTClient(bot)


# For a list of all built-in limiters, see https://arc.hypergonial.com/api_reference/utils/hooks/limiters/
@client.include
@arc.with_hook(arc.channel_limiter(10.0, 2))  # Limit the command to 2 uses every 10 seconds per channel.
@arc.slash_command("ping", "Pong!")
async def ping(ctx: arc.RESTContext) -> None:
    await ctx.respond("Pong!")


# You should be prepared to handle the `UnderCooldownError` exception, as it will be raised when the limiter is exhausted.
@ping.set_error_handler
async def ping_error_handler(ctx: arc.RESTContext, error: Exception) -> None:
    if isinstance(error, arc.UnderCooldownError):
        await ctx.respond(f"Command is on cooldown! Try again in `{error.retry_after}` seconds.")
    else:
        raise error


bot.run()



import hikari

import arc

# Context menus commands allow you to define commands that will appear in the right-click menu of a message or user.
# Related documentation guide: https://arc.hypergonial.com/guides/context_menu

bot = hikari.RESTBot("...")
client = arc.RESTClient(bot)

# Context-menu commands cannot be put in groups, and do not support options.
# Note that you can only define a MAXIMUM of 5 user and 5 message commands per bot.


@client.include
@arc.user_command("Say Hi")
# The user that was right-clicked is passed as a positional argument after the context.
async def hi_user(ctx: arc.RESTContext, user: hikari.User) -> None:
    await ctx.respond(f"Hey {user.mention}!")


@client.include
@arc.message_command("Say Hi")
# The message that was right-clicked is passed as a positional argument after the context.
async def hi_message(ctx: arc.RESTContext, message: hikari.Message) -> None:
    await ctx.respond(f"Hey {message.author.mention}!")


# Context-menu commands don't have the same naming restrictions as slash commands,
# (e.g. you can have spaces & uppercase letters),
# but two commands still cannot have the same name within the same category.

bot.run()



from typing import Any

import hikari

import arc

# Hooks are a way to execute common logic before or after a command is executed.
# Related documentation guide: https://arc.hypergonial.com/guides/hooks

bot = hikari.GatewayBot("...")
client = arc.GatewayClient(bot)


# Any function that takes a context as its first argument
# and returns None or HookResult is a valid hook
def my_hook(ctx: arc.Context[Any]) -> None:
    print(f"Hook was run on {ctx.command.name}!")


# Exceptions raised in hooks will abort the command
# The error can then be handled by an error handler
def my_check(ctx: arc.Context[Any]) -> None:
    if ctx.author.id != 1234567890:
        raise Exception("Unauthorized user tried to run command!")


# You can also use the HookResult class to return a result
# Note that abort=True silently cancels the command
# as opposed to raising an exception
def my_silent_check(ctx: arc.Context[Any]) -> arc.HookResult:
    if ctx.author.id != 1234567890:
        return arc.HookResult(abort=True)
    return arc.HookResult()


# Hooks can also be run after a command is executed
def cleanup_hook(ctx: arc.Context[Any]) -> None:
    print(f"Cleanup hook was run on {ctx.command.name}!")


# Hooks can be added to the client, plugins, groups, or individual commands
client.add_hook(my_check)


@client.include
@arc.with_post_hook(cleanup_hook)  # Or with '@arc.with_post_hook' to run the hook after the command
@arc.with_hook(my_hook)  # For commands, they can be added with the '@arc.with_hook' decorator
@arc.slash_command("ping", "Pong!")
async def ping(ctx: arc.GatewayContext) -> None:
    await ctx.respond("Pong!")


bot.run()



from __future__ import annotations

import abc
import asyncio
import typing as t

import attr
import hikari

from arc.abc.concurrency_limiting import ConcurrencyLimiterProto, HasConcurrencyLimiter
from arc.abc.error_handler import HasErrorHandler
from arc.abc.hookable import Hookable, HookResult
from arc.abc.limiter import LimiterProto
from arc.abc.option import OptionBase
from arc.context import AutodeferMode
from arc.errors import GlobalCommandPublishFailedError, MaxConcurrencyReachedError
from arc.internal.types import (
    BuilderT,
    ClientT,
    CommandCallbackT,
    ErrorHandlerCallbackT,
    HookT,
    PostHookT,
    ResponseBuilderT,
)
from arc.locale import CommandLocaleRequest

if t.TYPE_CHECKING:
    import typing_extensions as te

    from arc.abc.plugin import PluginBase
    from arc.context.base import Context


class CommandProto(t.Protocol):
    """A protocol for any command-like object. This includes commands, groups, subgroups, and subcommands."""

    name: str
    """The name of the command."""
    name_localizations: t.Mapping[hikari.Locale, str]
    """The name of the command in different locales."""

    @property
    @abc.abstractmethod
    def command_type(self) -> hikari.CommandType:
        """The type of command this object represents."""

    @property
    @abc.abstractmethod
    def qualified_name(self) -> t.Sequence[str]:
        """The fully qualified name of this command."""

    @property
    @abc.abstractmethod
    def display_name(self) -> str:
        """The display name of this command. This is what is shown in the Discord client.

        !!! note
            Slash commands can also be mentioned, see [SlashCommand.make_mention][arc.command.SlashCommand.make_mention].
        """


class CallableCommandProto(CommandProto, t.Protocol[ClientT]):
    """A protocol for any command-like object that can be called directly.

    This includes commands and subcommands, but not groups or subgroups.
    """

    name: str
    """The name of the command."""
    name_localizations: t.Mapping[hikari.Locale, str]
    """The name of the command in different locales."""

    callback: CommandCallbackT[ClientT]
    """The callback to invoke when this command is called."""

    @property
    @abc.abstractmethod
    def command_type(self) -> hikari.CommandType:
        """The type of command this object represents."""

    @property
    @abc.abstractmethod
    def qualified_name(self) -> t.Sequence[str]:
        """The fully qualified name of this command."""

    @property
    @abc.abstractmethod
    def display_name(self) -> str:
        """The display name of this command. This is what is shown in the Discord client.

        !!! note
            Slash commands can also be mentioned, see [SlashCommand.make_mention][arc.command.SlashCommand.make_mention].
        """

    @property
    @abc.abstractmethod
    def hooks(self) -> t.MutableSequence[HookT[ClientT]]:
        """The pre-execution hooks for this command."""

    @property
    @abc.abstractmethod
    def post_hooks(self) -> t.MutableSequence[PostHookT[ClientT]]:
        """The post-execution hooks for this command."""

    @abc.abstractmethod
    def reset_all_limiters(self, context: Context[ClientT]) -> None:
        """Reset all limiter hooks for this command.

        Parameters
        ----------
        context : Context
            The context to reset the limiters for.
        """

    @abc.abstractmethod
    async def __call__(self, ctx: Context[ClientT], *args: t.Any, **kwargs: t.Any) -> None:
        """Call the callback of the command with the given context and arguments.

        Parameters
        ----------
        ctx : ContextT
            The context to invoke this command with.
        args : tuple[Any]
            The positional arguments to pass to the callback.
        kwargs : dict[str, Any]
            The keyword arguments to pass to the callback.
        """

    @abc.abstractmethod
    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> None | asyncio.Future[ResponseBuilderT]:
        """Invoke this command with the given context.

        Parameters
        ----------
        interaction : hikari.CommandInteraction
            The interaction to invoke this command with.
        args : tuple[Any]
            The positional arguments to pass to the callback.
        kwargs : dict[str, Any]
            The keyword arguments to pass to the callback.

        Returns
        -------
        None | asyncio.Future[ResponseBuilderT]
            If this is a REST client, returns the response builder.

        Raises
        ------
        RuntimeError
            If this command has not been added to a client.
        """

    @abc.abstractmethod
    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        """Handle an exception that occurred while invoking this command.

        Parameters
        ----------
        ctx : Context
            The context that the exception occurred in.
        exc : Exception
            The exception that occurred.
        """

    @abc.abstractmethod
    def _resolve_hooks(self) -> t.Sequence[HookT[ClientT]]:
        """Resolve all pre-execution hooks that apply to this object."""

    @abc.abstractmethod
    def _resolve_post_hooks(self) -> t.Sequence[PostHookT[ClientT]]:
        """Resolve all post-execution hooks that apply to this object."""

    @abc.abstractmethod
    def _resolve_concurrency_limiter(self) -> ConcurrencyLimiterProto[ClientT] | None:
        """Resolve the concurrency limiter for this object."""


@t.final
@attr.define(slots=True, kw_only=True, weakref_slot=False)
class _CommandSettings:
    """All the command settings that need to propagate and be inherited."""

    autodefer: AutodeferMode | hikari.UndefinedType
    default_permissions: hikari.Permissions | hikari.UndefinedType
    is_nsfw: bool | hikari.UndefinedType
    integration_types: t.Sequence[hikari.ApplicationIntegrationType] | hikari.UndefinedType
    invocation_contexts: t.Sequence[hikari.ApplicationContextType] | hikari.UndefinedType

    def apply(self, other: te.Self) -> te.Self:
        """Apply 'other' to this, copying all the non-undefined settings to it."""
        return type(self)(
            autodefer=other.autodefer if other.autodefer is not hikari.UNDEFINED else self.autodefer,
            default_permissions=other.default_permissions
            if other.default_permissions is not hikari.UNDEFINED
            else self.default_permissions,
            is_nsfw=other.is_nsfw if other.is_nsfw is not hikari.UNDEFINED else self.is_nsfw,
            integration_types=other.integration_types
            if other.integration_types is not hikari.UNDEFINED
            else self.integration_types,
            invocation_contexts=other.invocation_contexts
            if other.invocation_contexts is not hikari.UNDEFINED
            else self.invocation_contexts,
        )

    @classmethod
    def default(cls) -> te.Self:
        """Get the default command settings."""
        return cls(
            autodefer=AutodeferMode.ON,
            default_permissions=hikari.UNDEFINED,
            is_nsfw=False,
            integration_types=[hikari.ApplicationIntegrationType.GUILD_INSTALL],
            invocation_contexts=[
                hikari.ApplicationContextType.GUILD,
                hikari.ApplicationContextType.BOT_DM,
                hikari.ApplicationContextType.PRIVATE_CHANNEL,
            ],
        )


@attr.define(slots=True, kw_only=True)
class CommandBase(
    HasErrorHandler[ClientT], Hookable[ClientT], HasConcurrencyLimiter[ClientT], t.Generic[ClientT, BuilderT]
):
    """An abstract base class for all application commands.

    This notably does not include subcommands & subgroups as those are in reality options.
    """

    name: str
    """The name of this command."""

    guilds: t.Sequence[hikari.Snowflake] | hikari.UndefinedType = hikari.UNDEFINED
    """The guilds this command is available in."""

    _autodefer: AutodeferMode | hikari.UndefinedType = attr.field(default=hikari.UNDEFINED, alias="autodefer")
    """If ON, this command will be automatically deferred if it takes longer than 2 seconds to respond."""

    _integration_types: t.Sequence[hikari.ApplicationIntegrationType] | hikari.UndefinedType = attr.field(
        default=hikari.UNDEFINED, alias="integration_types"
    )
    """A list of integration types where the command can be installed in."""

    _invocation_contexts: t.Sequence[hikari.ApplicationContextType] | hikari.UndefinedType = attr.field(
        default=hikari.UNDEFINED, alias="invocation_contexts"
    )
    """A list of context types where the command can be used in."""

    _default_permissions: hikari.Permissions | hikari.UndefinedType = attr.field(
        default=hikari.UNDEFINED, alias="default_permissions"
    )
    """The default permissions for this command.
    Keep in mind that guild administrators can change this, it should only be used to provide safe defaults."""

    _is_nsfw: bool | hikari.UndefinedType = attr.field(default=hikari.UNDEFINED, alias="is_nsfw")
    """Whether this command is NSFW. If true, the command will only be available in NSFW channels."""

    name_localizations: t.Mapping[hikari.Locale, str] = attr.field(factory=dict)
    """The localizations for this command's name."""

    _instances: dict[hikari.Snowflake | None, hikari.PartialCommand] = attr.field(factory=dict)
    """A mapping of guild IDs to command instances. None corresponds to the global instance, if any."""

    _client: ClientT | None = attr.field(init=False, default=None)
    """The client that is handling this command."""

    _plugin: PluginBase[ClientT] | None = attr.field(init=False, default=None)
    """The plugin that this command belongs to, if any."""

    _error_handler: ErrorHandlerCallbackT[ClientT] | None = attr.field(init=False, default=None)
    """The error handler for this command."""

    _concurrency_limiter: ConcurrencyLimiterProto[ClientT] | None = attr.field(init=False, default=None)
    """The concurrency limiter for this command."""

    _hooks: list[HookT[ClientT]] = attr.field(init=False, factory=list)
    """The pre-execution hooks for this command."""

    _post_hooks: list[PostHookT[ClientT]] = attr.field(init=False, factory=list)
    """The post-execution hooks for this command."""

    @property
    def error_handler(self) -> ErrorHandlerCallbackT[ClientT] | None:
        """The error handler for this command."""
        return self._error_handler

    @error_handler.setter
    def error_handler(self, callback: ErrorHandlerCallbackT[ClientT] | None) -> None:
        """Set the error handler for this command."""
        self._error_handler = callback

    @property
    def concurrency_limiter(self) -> ConcurrencyLimiterProto[ClientT] | None:
        """The concurrency limiter for this command."""
        return self._concurrency_limiter

    @concurrency_limiter.setter
    def concurrency_limiter(self, limiter: ConcurrencyLimiterProto[ClientT] | None) -> None:
        """Set the concurrency limiter for this command."""
        self._concurrency_limiter = limiter

    @property
    def hooks(self) -> t.MutableSequence[HookT[ClientT]]:
        """The pre-execution hooks for this command."""
        return self._hooks

    @property
    def post_hooks(self) -> t.MutableSequence[PostHookT[ClientT]]:
        """The post-execution hooks for this command."""
        return self._post_hooks

    @property
    @abc.abstractmethod
    def command_type(self) -> hikari.CommandType:
        """The type of command this object represents."""

    @property
    @abc.abstractmethod
    def qualified_name(self) -> t.Sequence[str]:
        """The fully qualified name of this command."""

    @property
    def client(self) -> ClientT:
        """The client that is handling this command."""
        if self._client is None:
            raise RuntimeError(
                f"Command '{self.display_name}' was not included in a client, '{type(self).__name__}.client' cannot be accessed until it is included in a client."
            )
        return self._client

    @property
    def plugin(self) -> PluginBase[ClientT] | None:
        """The plugin that this command belongs to, if any."""
        return self._plugin

    @property
    def autodefer(self) -> AutodeferMode:
        """The resolved autodefer configuration for this command."""
        settings = self._resolve_settings()
        return settings.autodefer if settings.autodefer is not hikari.UNDEFINED else AutodeferMode.ON

    @property
    def integration_types(self) -> t.Sequence[hikari.ApplicationIntegrationType]:
        """A list of integration types where the command can be installed in."""
        settings = self._resolve_settings()
        return (
            settings.integration_types
            if settings.integration_types is not hikari.UNDEFINED
            else [hikari.ApplicationIntegrationType.GUILD_INSTALL]
        )

    @property
    def invocation_contexts(self) -> t.Sequence[hikari.ApplicationContextType]:
        """A list of context types where the command can be used in."""
        settings = self._resolve_settings()
        return (
            settings.invocation_contexts
            if settings.invocation_contexts is not hikari.UNDEFINED
            else [
                hikari.ApplicationContextType.GUILD,
                hikari.ApplicationContextType.BOT_DM,
                hikari.ApplicationContextType.PRIVATE_CHANNEL,
            ]
        )

    @property
    def default_permissions(self) -> hikari.Permissions | hikari.UndefinedType:
        """The resolved default permissions for this command."""
        return self._resolve_settings().default_permissions

    @property
    def is_nsfw(self) -> bool:
        """Whether this command is NSFW. If true, the command will only be available in NSFW channels."""
        settings = self._resolve_settings()
        return settings.is_nsfw if settings.is_nsfw is not hikari.UNDEFINED else False

    @property
    def instances(self) -> t.Mapping[hikari.Snowflake | None, hikari.PartialCommand]:
        """A mapping of guild IDs to command instances. None corresponds to the global instance, if any."""
        return self._instances

    @property
    def display_name(self) -> str:
        """The display name of this command. This is what is shown in the Discord client.

        !!! note
            Slash commands can also be mentioned, see [SlashCommand.make_mention][arc.command.SlashCommand.make_mention].
        """
        return self.name

    def _register_instance(
        self,
        instance: hikari.PartialCommand,
        guild: hikari.SnowflakeishOr[hikari.PartialGuild] | hikari.UndefinedType = hikari.UNDEFINED,
    ) -> None:
        self._instances[hikari.Snowflake(guild) if guild else None] = instance

    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        try:
            if self.error_handler is not None:
                await ctx._injection_ctx.call_with_async_di(self.error_handler, ctx, exc)
            else:
                raise exc
        except Exception as exc:
            if self.plugin:
                await self.plugin._handle_exception(ctx, exc)
            else:
                await self.client._on_error(ctx, exc)

    def _resolve_settings(self) -> _CommandSettings:
        """Resolve all settings that apply to this command."""
        if self._plugin:
            settings = self._plugin._resolve_settings()
        elif self._client:
            settings = self._client._cmd_settings
        else:
            settings = _CommandSettings.default()

        return settings.apply(
            _CommandSettings(
                autodefer=self._autodefer,
                default_permissions=self._default_permissions,
                is_nsfw=self._is_nsfw,
                integration_types=self._integration_types,
                invocation_contexts=self._invocation_contexts,
            )
        )

    def _resolve_concurrency_limiter(self) -> ConcurrencyLimiterProto[ClientT] | None:
        """Resolve the concurrency limiter for this command."""
        if self._concurrency_limiter is not None:
            return self._concurrency_limiter

        if self._plugin:
            return self._plugin._resolve_concurrency_limiter()

        if self._client:
            return self._client._concurrency_limiter

        return None

    def _resolve_hooks(self) -> list[HookT[ClientT]]:
        upstream_hooks = self.plugin._resolve_hooks() if self.plugin else self.client._hooks
        return upstream_hooks + self._hooks

    def _resolve_post_hooks(self) -> list[PostHookT[ClientT]]:
        upstream_hooks = self.plugin._resolve_post_hooks() if self.plugin else self.client._post_hooks
        return upstream_hooks + self._post_hooks

    async def publish(self, guild: hikari.SnowflakeishOr[hikari.PartialGuild] | None = None) -> hikari.PartialCommand:
        """Publish this command to the given guild, or globally if no guild is provided.

        Parameters
        ----------
        guild : hikari.Snowflakeish | None
            The guild to publish this command to. If None, publish globally.

        Returns
        -------
        hikari.PartialCommand
            The published command.
        """
        if self.client.application is None:
            raise RuntimeError("Cannot publish command without a client.")

        kwargs = self._to_dict()
        try:
            if self.command_type is hikari.CommandType.SLASH:
                created = await self.client.app.rest.create_slash_command(self.client.application, **kwargs)
            else:
                created = await self.client.app.rest.create_context_menu_command(
                    self.client.application, type=self.command_type, **kwargs
                )
        except Exception as e:
            raise GlobalCommandPublishFailedError(self, f"Failed to publish command '{self.display_name}'") from e

        self._instances[hikari.Snowflake(guild) if guild else None] = created

        return created

    async def unpublish(self, guild: hikari.SnowflakeishOr[hikari.PartialGuild] | None = None) -> None:
        """Unpublish this command from the given guild, or globally if no guild is provided.

        Parameters
        ----------
        guild : hikari.Snowflakeish | None
            The guild to unpublish this command from. If None, unpublish globally.
        """
        if command := self._instances.pop(hikari.Snowflake(guild) if guild else None, None):
            await command.delete()

    @abc.abstractmethod
    def _get_context(
        self, interaction: hikari.CommandInteraction, command: CallableCommandProto[ClientT]
    ) -> Context[ClientT]:
        """Create a context object from an interaction."""

    @abc.abstractmethod
    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> None | asyncio.Future[ResponseBuilderT]:
        """Invoke this command with the given interaction.

        Parameters
        ----------
        interaction : hikari.CommandInteraction
            The interaction to invoke this command with.
        args : tuple[Any]
            The positional arguments to pass to the callback.
        kwargs : dict[str, Any]
            The keyword arguments to pass to the callback.

        Returns
        -------
        None | asyncio.Future[ResponseBuilderT]
            If this is a REST client, returns the response builder.

        Raises
        ------
        RuntimeError
            If this command has not been added to a client.
        """

    def _to_dict(self) -> dict[str, t.Any]:
        return {
            "name": self.name,
            "nsfw": self.is_nsfw,
            "default_member_permissions": self.default_permissions,
            "name_localizations": self.name_localizations,
            "integration_types": self.integration_types,
            "invocation_contexts": self.invocation_contexts,
        }

    @abc.abstractmethod
    def _build(self, id: hikari.Snowflake | hikari.UndefinedType = hikari.UNDEFINED) -> BuilderT:
        """Create a builder out of this command.

        Parameters
        ----------
        id : hikari.Snowflake | hikari.UndefinedType
            The ID of the command, if it already exists.
            If provided, this will edit the existing command, otherwise it will create a new one.
        """

    def _client_include_hook(self, client: ClientT) -> None:
        """Called when the client requests the command be added to it."""
        self._client = client
        self.client._add_command(self)

    def _client_remove_hook(self, client: ClientT) -> None:
        """Called when the client requests the command be removed from it."""
        self.client._remove_command(self)
        self._client = None

    def _plugin_include_hook(self, plugin: PluginBase[ClientT]) -> None:
        """Called when the plugin requests the command be added to it."""
        self._plugin = plugin
        self._plugin._add_command(self)

    def _request_command_locale(self) -> None:
        """Request the locale for this command."""
        if self.name_localizations or self._client is None:
            return

        if not self._client._provided_locales or not self._client._command_locale_provider:
            return

        name_locales: dict[hikari.Locale, str] = {}

        for locale in self._client._provided_locales:
            request = CommandLocaleRequest(self, locale, self.name)
            resp = self._client._command_locale_provider(request)

            if resp.name is not None:
                name_locales[locale] = resp.name

        self.name_localizations = name_locales

    async def _handle_pre_hooks(self, command: CallableCommandProto[ClientT], ctx: Context[ClientT]) -> bool:
        """Handle all pre-execution hooks for a command.

        Returns
        -------
        bool
            Whether the command should be aborted.
        """
        aborted = False
        try:
            hooks = command._resolve_hooks()
            for hook in hooks:
                res = await ctx._injection_ctx.call_with_async_di(hook, ctx)

                res = t.cast("HookResult | None", res)

                if res and res._abort:
                    aborted = True
        except Exception as e:
            aborted = True
            await command._handle_exception(ctx, e)

        return aborted

    async def _handle_post_hooks(self, command: CallableCommandProto[ClientT], ctx: Context[ClientT]) -> None:
        """Handle all post-execution hooks for a command, and release the concurrency limiter if applicable."""
        try:
            post_hooks = command._resolve_post_hooks()
            for hook in post_hooks:
                await ctx._injection_ctx.call_with_async_di(hook, ctx)
        except Exception as e:
            await command._handle_exception(ctx, e)
        finally:
            if (limiter := command._resolve_concurrency_limiter()) is not None:
                limiter.release(ctx)

    async def _handle_callback(
        self, command: CallableCommandProto[ClientT], ctx: Context[ClientT], *args: t.Any, **kwargs: t.Any
    ) -> None:
        """Handle the callback of a command. Invoke all hooks and the callback, and handle any exceptions."""
        # If hook aborted, stop invocation

        injection_ctx = await self.client._create_overriding_ctx_for_command(ctx)
        ctx._injection_ctx = injection_ctx

        max_concurrency = command._resolve_concurrency_limiter()

        if max_concurrency is not None and max_concurrency.is_exhausted(ctx):
            return await command._handle_exception(
                ctx, MaxConcurrencyReachedError(max_concurrency, max_concurrency.limit)
            )

        try:
            if max_concurrency is not None:
                await max_concurrency.acquire(ctx)

            if await self._handle_pre_hooks(command, ctx):
                return

            await injection_ctx.call_with_async_di(command.callback, ctx, *args, **kwargs)

        except Exception as e:
            ctx._has_command_failed = True
            await command._handle_exception(ctx, e)
        finally:
            # This also releases the concurrency limiter
            await self._handle_post_hooks(command, ctx)


@attr.define(slots=True, kw_only=True)
class CallableCommandBase(CommandBase[ClientT, BuilderT], CallableCommandProto[ClientT]):
    """A top-level command that can be called directly. Note that this does not include subcommands, as those are options."""

    callback: CommandCallbackT[ClientT]
    """The callback to invoke when this command is called."""

    _invoke_task: asyncio.Task[t.Any] | None = attr.field(init=False, default=None, repr=False)

    def reset_all_limiters(self, context: Context[ClientT]) -> None:
        """Reset all limiter hooks for this command.

        Parameters
        ----------
        context : Context
            The context to reset the limiters for.
        """
        limiters: t.Generator[LimiterProto[ClientT], None, None] = (
            lim for lim in self._resolve_hooks() if isinstance(lim, LimiterProto)
        )
        for limiter in limiters:
            limiter.reset(context)

    async def __call__(self, ctx: Context[ClientT], *args: t.Any, **kwargs: t.Any) -> None:
        await self.callback(ctx, *args, **kwargs)

    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> None | asyncio.Future[ResponseBuilderT]:
        ctx = self._get_context(interaction, self)
        if self.autodefer.should_autodefer:
            ctx._start_autodefer(self.autodefer)
        self._invoke_task = asyncio.create_task(self._handle_callback(self, ctx, *args, **kwargs))
        if self.client.is_rest:
            return ctx._resp_builder


ParentT = t.TypeVar("ParentT")


@attr.define(slots=True, kw_only=True)
class SubCommandBase(
    OptionBase[ClientT],
    HasErrorHandler[ClientT],
    Hookable[ClientT],
    HasConcurrencyLimiter[ClientT],
    t.Generic[ClientT, ParentT],
):
    """An abstract base class for all slash subcommands and subgroups."""

    _error_handler: ErrorHandlerCallbackT[ClientT] | None = attr.field(default=None, init=False)

    _concurrency_limiter: ConcurrencyLimiterProto[ClientT] | None = attr.field(default=None, init=False)

    _hooks: list[HookT[ClientT]] = attr.field(factory=list, init=False)

    _post_hooks: list[PostHookT[ClientT]] = attr.field(factory=list, init=False)

    _parent: ParentT | None = attr.field(default=None, init=False, alias="parent")
    """The parent of this subcommand or subgroup."""

    @property
    @abc.abstractmethod
    def display_name(self) -> str:
        """The display name of this command. This is what is shown in the Discord client.

        !!! note
            Slash commands can also be mentioned, see [SlashCommand.make_mention][arc.command.SlashCommand.make_mention].
        """

    @property
    def error_handler(self) -> ErrorHandlerCallbackT[ClientT] | None:
        """The error handler for this object."""
        return self._error_handler

    @error_handler.setter
    def error_handler(self, callback: ErrorHandlerCallbackT[ClientT] | None) -> None:
        """Set the error handler for this object."""
        self._error_handler = callback

    @property
    def concurrency_limiter(self) -> ConcurrencyLimiterProto[ClientT] | None:
        """The concurrency limiter for this object."""
        return self._concurrency_limiter

    @concurrency_limiter.setter
    def concurrency_limiter(self, limiter: ConcurrencyLimiterProto[ClientT] | None) -> None:
        """Set the concurrency limiter for this object."""
        self._concurrency_limiter = limiter

    @property
    def hooks(self) -> t.MutableSequence[HookT[ClientT]]:
        """The pre-execution hooks for this object."""
        return self._hooks

    @property
    def post_hooks(self) -> t.MutableSequence[PostHookT[ClientT]]:
        """The post-execution hooks for this object."""
        return self._post_hooks

    @property
    def parent(self) -> ParentT:
        """The parent of this subcommand or subgroup."""
        if self._parent is None:
            raise RuntimeError(
                f"Subcommand '{self.name}' was not included in a parent, '{type(self).__name__}.parent' cannot be accessed until it is included in a parent."
            )
        return self._parent

    def reset_all_limiters(self, context: Context[ClientT]) -> None:
        """Reset all limiter hooks for this command.

        Parameters
        ----------
        context : Context
            The context to reset the limiters for.
        """
        limiters: t.Generator[LimiterProto[ClientT], None, None] = (
            lim for lim in self._resolve_hooks() if isinstance(lim, LimiterProto)
        )
        for limiter in limiters:
            limiter.reset(context)


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

bot = hikari.GatewayBot("...")
client = arc.GatewayClient(bot)

# Related documentation guide: https://arc.hypergonial.com/guides/loops


# Create a loop out of a function
# This will call the function every 10 seconds when started
@arc.utils.interval_loop(seconds=10.0)
async def loopy_loop(value: int) -> None:
    print(value)


async def another_loop() -> None:
    print("Every 60 seconds, a minute passes")


# You may also create a loop using the class directly
loop = arc.utils.IntervalLoop(another_loop, seconds=60.0)


@client.add_startup_hook
async def startup(client: arc.GatewayClient) -> None:
    # Start the loop by passing all the parameters it needs
    loopy_loop.start(value=10)

    # The other loop has no parameters, so you can just start it
    loop.start()


bot.run()



from __future__ import annotations

import typing as t

import hikari

from arc.abc.hookable import HookResult
from arc.context import Context  # noqa: TC001 Needed for DI to work
from arc.errors import (
    BotMissingPermissionsError,
    DMOnlyError,
    GuildOnlyError,
    InvokerMissingPermissionsError,
    NotOwnerError,
)


def guild_only(ctx: Context[t.Any]) -> HookResult:
    """A pre-execution hook that aborts the execution of a command if it is invoked outside of a guild.

    Example
    --------
    ```py
    @arc.with_hook(arc.guild_only)
    ```

    Raises
    ------
    GuildOnlyError
        If the command is invoked outside of a guild.
    """
    if ctx.guild_id is None:
        raise GuildOnlyError("This command can only be used in a guild.")
    return HookResult()


def dm_only(ctx: Context[t.Any]) -> HookResult:
    """A pre-execution hook that aborts the execution of a command if it is invoked outside of a DM.

    Example
    --------
    ```py
    @arc.with_hook(arc.dm_only)
    ```

    Raises
    ------
    DMOnlyError
        If the command is invoked outside of a DM.
    """
    if ctx.guild_id is not None:
        raise DMOnlyError("This command can only be used in a DM.")
    return HookResult()


def owner_only(ctx: Context[t.Any]) -> HookResult:
    """A pre-execution hook that aborts the execution of a command if it is invoked by a non-owner.

    Example
    --------
    ```py
    @arc.with_hook(arc.owner_only)
    ```

    Raises
    ------
    NotOwnerError
        If the command is invoked by a non-owner.
    """
    if ctx.author.id not in ctx.client.owner_ids:
        raise NotOwnerError("This command can only be used by the application owners.")
    return HookResult()


def _has_permissions(ctx: Context[t.Any], perms: hikari.Permissions) -> HookResult:
    """Check if the invoker has the specified permissions."""
    if ctx.member is None:
        raise GuildOnlyError("This command can only be used in a guild.")

    missing_perms = ~ctx.member.permissions & perms

    if missing_perms is not hikari.Permissions.NONE:
        raise InvokerMissingPermissionsError(
            missing_perms, f"Invoker is missing '{missing_perms}' permissions to run this command."
        )

    return HookResult()


def has_permissions(perms: hikari.Permissions) -> t.Callable[[Context[t.Any]], HookResult]:
    """A pre-execution hook that aborts the execution of a command if the invoker is missing the specified permissions
    in the channel the command was invoked in.

    Parameters
    ----------
    perms : hikari.Permissions
        The permissions to check for.

    Raises
    ------
    GuildOnlyError
        If the command is invoked outside of a guild.
    InvokerMissingPermissionsError
        If the invoker is missing some of the specified permissions.

    Example
    --------
    ```py
    @arc.with_hook(arc.has_permissions(hikari.Permissions.MANAGE_CHANNELS | hikari.Permissions.MANAGE_GUILD))
    ```
    You can combine permissions with the bitwise OR operator (`|`).

    !!! note
        This hook requires the command to be invoked in a guild, and implies the [`guild_only`][arc.utils.hooks.guild_only] hook.
    """
    return lambda ctx: _has_permissions(ctx, perms)


def _bot_has_permissions(ctx: Context[t.Any], perms: hikari.Permissions) -> HookResult:
    """Check if the bot has the specified permissions."""
    missing_perms = ~ctx.app_permissions & perms

    if missing_perms is not hikari.Permissions.NONE:
        raise BotMissingPermissionsError(
            missing_perms, f"Bot is missing '{missing_perms}' permissions to run this command."
        )

    return HookResult()


def bot_has_permissions(perms: hikari.Permissions) -> t.Callable[[Context[t.Any]], HookResult]:
    """A pre-execution hook that aborts the execution of a command if the bot is missing the specified permissions
    in the channel the command was invoked in.

    Parameters
    ----------
    perms : hikari.Permissions
        The permissions to check for.

    Raises
    ------
    GuildOnlyError
        If the command is invoked outside of a guild.
    BotMissingPermissionsError
        If the bot is missing some of the specified permissions.

    Example
    --------
    ```py
    @arc.with_hook(arc.bot_has_permissions(hikari.Permissions.MANAGE_CHANNELS | hikari.Permissions.MANAGE_GUILD))
    ```
    You can combine permissions with the bitwise OR operator (`|`).

    !!! note
        This hook requires the command to be invoked in a guild, and implies the [`guild_only`][arc.utils.hooks.guild_only] hook.
    """
    return lambda ctx: _bot_has_permissions(ctx, perms)


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t

import attr
import hikari

from arc.abc.option import CommandOptionBase, OptionParams, OptionType
from arc.internal.types import ClientT

if t.TYPE_CHECKING:
    import typing_extensions as te

__all__ = ("ChannelOption", "ChannelParams")


@t.final
class ChannelParams(OptionParams[hikari.PartialChannel]):
    """The parameters for a channel option.
    The channel types are inferred from the type hint.

    !!! example
        An option of type `hikari.TextableGuildChannel` will resolve to `hikari.ChannelType.GUILD_PUBLIC_THREAD |
        hikari.ChannelType.GUILD_PRIVATE_THREAD |
        hikari.ChannelType.GUILD_NEWS_THREAD`.

    Parameters
    ----------
    description : str
        The description of the option
    name : str
        The name of the option. If not provided, the name of the parameter will be used.
    name_localizations : Mapping[hikari.Locale, str]
        The name of the option in different locales
    description_localizations : Mapping[hikari.Locale, str]
        The description of the option in different locales
    """

    __slots__: t.Sequence[str] = ()


@attr.define(slots=True, kw_only=True)
class ChannelOption(CommandOptionBase[hikari.PartialChannel, ClientT, ChannelParams]):
    """A slash command option that represents a channel.

    ??? hint
        To add an option of this type to your command, add an argument to your command function with the following type hint:
        ```py
        opt_name: arc.Option[hikari.TextableGuildChannel, ChannelParams(...)]
        ```

        **Note:** The type of the channel determines the channel types that this option can be used with.
    """

    channel_types: t.Sequence[hikari.ChannelType] | None = None
    """The channel types that the option can be."""

    @property
    def option_type(self) -> OptionType:
        return OptionType.CHANNEL

    @classmethod
    def _from_params(
        cls, *, name: str, arg_name: str, is_required: bool, params: ChannelParams, **kwargs: t.Any
    ) -> te.Self:
        channel_types = kwargs.pop("channel_types")
        return cls(
            name=name,
            arg_name=arg_name,
            description=params.description,
            is_required=is_required,
            name_localizations=params.name_localizations,
            description_localizations=params.description_localizations,
            channel_types=channel_types,
            **kwargs,
        )

    def _to_dict(self) -> dict[str, t.Any]:
        return {**super()._to_dict(), "channel_types": self.channel_types or None}


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t

import hikari

from arc.internal.types import GatewayClientT

if t.TYPE_CHECKING:
    from arc.context import Context

__all__ = ("ArcEvent", "CommandErrorEvent", "StartedEvent", "StoppingEvent")


class ArcEvent(hikari.Event):
    """Base class for all Arc events."""

    __slots__: t.Sequence[str] = ()


class StartedEvent(ArcEvent, t.Generic[GatewayClientT]):
    """Event dispatched when the client has started.

    This is notably different from hikari's `StartedEvent`
    because it fires after all command syncing was completed and the client has finished its startup process.
    """

    __slots__: t.Sequence[str] = ("_client",)

    def __init__(self, client: GatewayClientT) -> None:
        self._client = client

    @property
    def client(self) -> GatewayClientT:
        """The client instance that started."""
        return self._client

    @property
    def app(self) -> hikari.RESTAware:
        """App instance for this application."""
        return self._client.app


class StoppingEvent(ArcEvent, t.Generic[GatewayClientT]):
    """Event dispatched when the client is stopping.

    This event is fired after the shutdown hook has been processed.
    """

    __slots__: t.Sequence[str] = ("_client",)

    def __init__(self, client: GatewayClientT) -> None:
        self._client = client

    @property
    def client(self) -> GatewayClientT:
        """The client instance that is stopping."""
        return self._client

    @property
    def app(self) -> hikari.RESTAware:
        """App instance for this application."""
        return self._client.app


class CommandErrorEvent(ArcEvent, t.Generic[GatewayClientT]):
    """Event dispatched when a command raises an exception that is not handled by any error handlers.

    !!! warning
        Creating any listeners for this event will disable the client error handler completely.
    """

    __slots__: t.Sequence[str] = ("_client", "_context", "_exception")

    def __init__(self, client: GatewayClientT, context: Context[GatewayClientT], exception: Exception) -> None:
        self._context = context
        self._client = client
        self._exception = exception

    @property
    def client(self) -> GatewayClientT:
        """The client instance that raised the exception."""
        return self._client

    @property
    def app(self) -> hikari.RESTAware:
        """App instance for this application."""
        return self._client.app

    @property
    def context(self) -> Context[GatewayClientT]:
        """The invocation context that raised the exception."""
        return self._context

    @property
    def exception(self) -> Exception:
        """The exception that was raised."""
        return self._exception



import typing as t

import hikari
import pytest

import arc

bot = hikari.GatewayBot("...", banner=None)
client = arc.GatewayClient(bot)


class Dummy:
    def __init__(self, val: int):
        self.val = val


client.set_type_dependency(Dummy, Dummy(0))


@client.inject_dependencies
def dummy_callback(dummy: Dummy = arc.inject()) -> int:
    dummy.val += 1
    return dummy.val


@client.inject_dependencies
async def async_dummy_callback(dummy: Dummy = arc.inject()) -> int:
    dummy.val += 1
    return dummy.val


@client.inject_dependencies
def injects_client(client: arc.abc.Client[t.Any] = arc.inject()) -> arc.abc.Client[t.Any]:
    return client


@client.inject_dependencies  # type: ignore
def injects_client_improperly_typed(client: arc.abc.Client = arc.inject()) -> arc.abc.Client:  # type: ignore
    return client  # type: ignore


@client.inject_dependencies
def injects_exact_client(client: arc.GatewayClient = arc.inject()) -> arc.GatewayClient:
    return client


injector = arc.Injector()
injector.set_type_dependency(Dummy, Dummy(0))

manually_injected_client = arc.GatewayClient(bot, injector=injector)


@manually_injected_client.inject_dependencies
def manual_dummy_callback(dummy: Dummy = arc.inject()) -> int:
    dummy.val += 1
    return dummy.val


@pytest.mark.asyncio
async def test_inject() -> None:
    assert dummy_callback() == 1
    assert await async_dummy_callback() == 2
    assert dummy_callback() == 3
    assert injects_client() is client
    assert injects_client_improperly_typed() is client
    assert injects_exact_client() is client

    assert manual_dummy_callback() == 1



from __future__ import annotations

import typing as t

import attr
import hikari

from arc.abc.option import ConverterOption, OptionParams, OptionType
from arc.errors import OptionConverterFailureError
from arc.internal.types import ClientT

if t.TYPE_CHECKING:
    import typing_extensions as te


__all__ = ("EmojiOption", "EmojiParams")


@t.final
class EmojiParams(OptionParams[hikari.Color]):
    """The parameters for an emoji option.

    Parameters
    ----------
    description : str
        The description of the option
    name : str
        The name of the option. If not provided, the name of the parameter will be used.
    name_localizations : Mapping[hikari.Locale, str] | None
        The name of the option in different locales
    description_localizations : Mapping[hikari.Locale, str] | None
        The description of the option in different locales
    """

    __slots__ = ()


@attr.define(slots=True, kw_only=True)
class EmojiOption(ConverterOption[hikari.Emoji, ClientT, EmojiParams, str]):
    """A slash command option that represents an emoji.

    ??? hint
        To add an option of this type to your command, add an argument to your command function with the following type hint:
        ```py
        opt_name: arc.Option[hikari.Emoji, EmojiParams(...)]
        ```
    """

    @property
    def option_type(self) -> OptionType:
        return OptionType.EMOJI

    def _convert_value(self, value: str) -> hikari.Emoji:
        try:
            return hikari.Emoji.parse(value)
        except ValueError as exc:
            raise OptionConverterFailureError(
                self, value, f"Option '{self.name}' expected a valid emoji, got {value!r}."
            ) from exc

    @classmethod
    def _from_params(
        cls, *, name: str, arg_name: str, is_required: bool, params: EmojiParams, **kwargs: t.Any
    ) -> te.Self:
        return cls(
            name=name,
            arg_name=arg_name,
            description=params.description,
            is_required=is_required,
            name_localizations=params.name_localizations,
            description_localizations=params.description_localizations,
            **kwargs,
        )


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

bot = hikari.GatewayBot("...")
# Set the locales that the client will request in the provider callbacks.
client = arc.GatewayClient(bot, provided_locales=[hikari.Locale.EN_US, hikari.Locale.ES_ES])

# These are just examples, you can provide localizations from anywhere you want.
COMMAND_LOCALES = {
    "hi": {
        hikari.Locale.EN_US: {"name": "hi", "description": "Say hi to someone!"},
        hikari.Locale.ES_ES: {"name": "hola", "description": "¡Saluda a alguien!"},
    }
}

OPTION_LOCALES = {
    "hi": {
        "user": {
            hikari.Locale.EN_US: {"name": "user", "description": "The user to say hi to."},
            hikari.Locale.ES_ES: {"name": "usuario", "description": "El usuario al que saludar."},
        }
    }
}

CUSTOM_LOCALES = {"say_hi": {hikari.Locale.EN_US: "Hey {user}!", hikari.Locale.ES_ES: "¡Hola {user}!"}}


# This callback will be called when a command needs to be localized.
# The request includes the command and the locale to be used.
@client.set_command_locale_provider
def command_locale_provider(request: arc.CommandLocaleRequest) -> arc.LocaleResponse:
    return arc.LocaleResponse(**COMMAND_LOCALES[request.name][request.locale])


# This callback will be called when an option needs to be localized.
# The request includes the option and the locale to be used.
@client.set_option_locale_provider
def option_locale_provider(request: arc.OptionLocaleRequest) -> arc.LocaleResponse:
    return arc.LocaleResponse(**OPTION_LOCALES[request.command.name][request.name][request.locale])


# This callback will be called for every 'ctx.loc()' call.
# The '.key' attribute is the string that is passed to the 'ctx.loc()' method.
@client.set_custom_locale_provider
def custom_locale_provider(request: arc.CustomLocaleRequest) -> str:
    return CUSTOM_LOCALES[request.key][request.locale]


@client.include
@arc.slash_command(name="hi", description="Say hi to someone!")
async def hi_slash(
    ctx: arc.GatewayContext, user: arc.Option[hikari.User, arc.UserParams("The user to say hi to.")]
) -> None:
    # ctx.loc can be used to request custom localizations that are not related to commands or options.
    # Additional keyword arguments passed to the method will be passed to .format()
    await ctx.respond(ctx.loc("say_hi", user=user.mention))


# Note: 'ctx.loc()' will by default request the guild's locale, outside of guilds it will request the user's locale.
# You can pass use_guild=False to always request the user's locale instead.


bot.run()



import hikari

import arc


# Define a custom bot class that extends hikari.RESTBot
# Note that this is not strictly required, but is possible
class MyBot(hikari.RESTBot):
    def custom_method(self) -> str:
        return "I'm doing stuff!"


# You should inherit from arc.RESTClientBase
# and specify the bot class that your client will use
# If you do not have a custom bot class, you can use hikari.RESTBot here
class MyClient(arc.RESTClientBase[MyBot]):
    def custom_client_method(self) -> None:
        print(self.app.custom_method())


# Optional but recommended:
# Create type aliases for your context and plugin types
MyContext = arc.Context[MyClient]
MyPlugin = arc.RESTPluginBase[MyClient]

# Create your bot instance and client instance
bot = MyBot("...", banner=None)
client = MyClient(bot)


# Use in commands like normal
@client.include()
@arc.slash_command("test", "My command description")
async def my_command(ctx: MyContext) -> None:
    ctx.client.custom_client_method()
    await ctx.respond(ctx.client.app.custom_method())


bot.run()



import hikari

import arc

bot = hikari.GatewayBot("...", banner=None)
client = arc.GatewayClient(bot)


def test_ext_load() -> None:
    client.load_extension("extension")
    assert len(client.plugins) == 1
    assert client.plugins["test_plugin"].name == "test_plugin"
    assert len(client._slash_commands) == 4
    assert {"foo", "bar", "baz", "my_group"} == set(client._slash_commands.keys())



import abc
import asyncio
import datetime
import inspect
import sys
import traceback
import typing as t

__all__ = ("CronLoop", "IntervalLoop", "cron_loop", "interval_loop")

P = t.ParamSpec("P")


class _LoopBase(abc.ABC, t.Generic[P]):
    """An abstract base class for loops.

    Parameters
    ----------
    callback : t.Callable[P, t.Awaitable[None]]
        The coroutine to run at the specified interval.

    See Also
    --------
    - [`IntervalLoop`][arc.utils.loops.IntervalLoop]
    - [`CronLoop`][arc.utils.loops.CronLoop]
    """

    __slots__ = ("_coro", "_failed", "_run_on_start", "_stop_next", "_task")

    def __init__(self, callback: t.Callable[P, t.Awaitable[None]], *, run_on_start: bool = True) -> None:
        self._coro = callback
        self._task: asyncio.Task[None] | None = None
        self._failed: int = 0
        self._run_on_start: bool = run_on_start
        self._stop_next: bool = False

        if not inspect.iscoroutinefunction(self._coro):
            raise TypeError("Expected a coroutine function.")

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Call the underlying coroutine."""
        await self._coro(*args, **kwargs)

    @abc.abstractmethod
    def _get_next_run(self) -> float:
        """Get the amount of time to sleep until the next run, in seconds.
        This will be called after the coroutine has finished running.

        Returns
        -------
        float
            The number of seconds to wait before running the coroutine again.
        """

    async def _call_callback(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Call the callback and handle exceptions."""
        try:
            await self._coro(*args, **kwargs)
        except Exception as e:
            traceback_msg = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
            print(f"Loop encountered exception: {e}", file=sys.stderr)
            print(traceback_msg, file=sys.stderr)

            if self._failed < 3:
                self._failed += 1
            else:
                raise RuntimeError(f"Loop failed repeatedly, stopping it. Exception: {e}")

    async def _loopy_loop(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Main loop logic."""
        if self._run_on_start:
            await self._call_callback(*args, **kwargs)

        while not self._stop_next:
            await asyncio.sleep(self._get_next_run())
            await self._call_callback(*args, **kwargs)
        self.cancel()

    def _create_task(self, *args: P.args, **kwargs: P.kwargs) -> asyncio.Task[None]:
        """Create the task for the loop."""
        task = asyncio.create_task(self._loopy_loop(*args, **kwargs))
        task.add_done_callback(self._handle_result)
        return task

    def _handle_result(self, task: asyncio.Task[None]) -> None:
        """Handle the result of the task."""
        if task.cancelled():
            return

        try:
            task.result()
        except Exception as e:
            traceback_msg = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
            print(f"Loop encountered exception: {e}", file=sys.stderr)
            print(traceback_msg, file=sys.stderr)

    def start(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Start the loop at the specified interval.

        Parameters
        ----------
        args : P.args
            The positional arguments to pass to the coroutine every iteration.
        kwargs : P.kwargs
            The keyword arguments to pass to the coroutine every iteration.
        """
        if self._task and not self._task.done():
            raise RuntimeError("Task is already running!")

        self._task = self._create_task(*args, **kwargs)

    def cancel(self) -> None:
        """Cancel the loop.

        If you want to stop the loop gracefully, use `stop()` instead.
        """
        if not self._task:
            return

        self._task.cancel()
        self._task = None

    def stop(self) -> None:
        """Gracefully stop the loop. This will wait for the current iteration to finish."""
        if self._task and not self._task.done():
            self._stop_next = True


class IntervalLoop(_LoopBase[P]):
    """A simple interval loop that runs a coroutine at a specified interval.

    Parameters
    ----------
    callback : t.Callable[P, t.Awaitable[None]]
        The coroutine to run at the specified interval.
    seconds : float, optional
        The number of seconds to wait before running the coroutine again.
    minutes : float, optional
        The number of minutes to wait before running the coroutine again.
    hours : float, optional
        The number of hours to wait before running the coroutine again.
    days : float, optional
        The number of days to wait before running the coroutine again.
    run_on_start : bool, optional
        Whether to run the callback immediately after starting the loop.
        If set to false, the loop will wait for the specified interval before first running the callback.

    Raises
    ------
    ValueError
        If no interval is specified.
    TypeError
        If the passed function is not a coroutine function.

    Example
    --------
    ```py
    loop = IntervalLoop(my_coro, seconds=5)
    loop.start()
    ```

    You may also use the decorator [`@arc.utils.interval_loop`][arc.utils.loops.interval_loop] to
    create an [`IntervalLoop`][arc.utils.loops.IntervalLoop] from a coroutine function.
    """

    __slots__ = ("_sleep",)

    def __init__(
        self,
        callback: t.Callable[P, t.Awaitable[None]],
        *,
        seconds: float | None = None,
        minutes: float | None = None,
        hours: float | None = None,
        days: float | None = None,
        run_on_start: bool = True,
    ) -> None:
        super().__init__(callback, run_on_start=run_on_start)
        if not seconds and not minutes and not hours and not days:
            raise ValueError("At least one of 'seconds', 'minutes', 'hours' or 'days' must be not None.")
        else:
            seconds = seconds or 0
            minutes = minutes or 0
            hours = hours or 0
            days = hours or 0

        self._sleep: float = seconds + minutes * 60 + hours * 3600 + days * 24 * 3600

    def _get_next_run(self) -> float:
        return self._sleep

    def set_interval(
        self,
        *,
        seconds: float | None = None,
        minutes: float | None = None,
        hours: float | None = None,
        days: float | None = None,
    ):
        """Set a new specified interval.

        !!! note
            You need to restart the loop if you want these changes to take effect immediately.

        Parameters
        ----------
        seconds : float | None, optional
            The number of seconds to wait before running the coroutine again.
        minutes : float | None, optional
            The number of minutes to wait before running the coroutine again.
        hours : float | None, optional
            The number of hours to wait before running the coroutine again.
        days : float | None, optional
            The number of days to wait before running the coroutine again.

        Example
        --------
        ```py
        loop = IntervalLoop(my_coro, seconds=5)
        loop.start()
        loop.set_interval(seconds=10)
        loop.cancel()
        loop.start()
        ```
        """
        if not seconds and not minutes and not hours and not days:
            raise ValueError("At least one of 'seconds', 'minutes', 'hours' or 'days' must be not None.")
        else:
            seconds = seconds or 0
            minutes = minutes or 0
            hours = hours or 0
            days = hours or 0

        self._sleep: float = seconds + minutes * 60 + hours * 3600 + days * 24 * 3600


class CronLoop(_LoopBase[P]):
    """A simple interval loop that runs a coroutine at a specified interval.

    !!! warning
        To use this loop, you must install arc with the `cron` extra.

        ```sh
        pip install hikari-arc[cron]
        ```

    Parameters
    ----------
    callback : t.Callable[P, t.Awaitable[None]]
        The coroutine to run at the specified interval.
    cron_format : str
        The cron format to use. See https://en.wikipedia.org/wiki/Cron for more information.
    timezone : datetime.timezone
        The timezone to use for the cron loop. Defaults to UTC.

    Raises
    ------
    ImportError
        If the `croniter` package is not installed.
    croniter.CroniterBadCronError
        If the cron format is invalid.
    TypeError
        If the passed function is not a coroutine function.

    Example
    --------
    ```py
    loop = CronLoop(my_coro, "*/5 * * * *")
    loop.start()
    ```

    You may also use the decorator [`@arc.utils.cron_loop`][arc.utils.loops.cron_loop] to
    create a [`CronLoop`][arc.utils.loops.CronLoop] from a coroutine function.
    """

    __slots__ = ("_iter", "_tz")

    def __init__(
        self,
        callback: t.Callable[P, t.Awaitable[None]],
        cron_format: str,
        *,
        timezone: datetime.timezone = datetime.timezone.utc,
    ) -> None:
        super().__init__(callback, run_on_start=False)
        self._tz = timezone

        try:
            import croniter

            self._iter = croniter.croniter(cron_format)
        except ImportError:
            raise ImportError("Missing dependency for CronLoop: 'croniter'")

    def _get_next_run(self) -> float:
        return (
            self._iter.get_next(float, start_time=datetime.datetime.now(self._tz))
            - datetime.datetime.now(self._tz).timestamp()
        )


def interval_loop(
    *,
    seconds: float | None = None,
    minutes: float | None = None,
    hours: float | None = None,
    days: float | None = None,
    run_on_start: bool = True,
) -> t.Callable[[t.Callable[P, t.Awaitable[None]]], IntervalLoop[P]]:
    """A decorator to create an [`IntervalLoop`][arc.utils.loops.IntervalLoop] out of a coroutine function.

    Parameters
    ----------
    seconds : float, optional
        The number of seconds to wait before running the coroutine again.
    minutes : float, optional
        The number of minutes to wait before running the coroutine again.
    hours : float, optional
        The number of hours to wait before running the coroutine again.
    days : float, optional
        The number of days to wait before running the coroutine again.
    run_on_start : bool, optional
        Whether to run the callback immediately after starting the loop.
        If set to false, the loop will wait for the specified interval before first running the callback.

    Returns
    -------
    t.Callable[[t.Callable[P, t.Awaitable[None]]], IntervalLoop[P]]
        The decorator.

    Raises
    ------
    ValueError
        If no interval is specified.
    TypeError
        If the decorated function is not a coroutine function.

    Example
    --------
    ```py
    import arc

    # Run every 5 seconds.
    @arc.utils.interval_loop(seconds=5)
    async def my_loop():
        print("Hello, loop!")

    # Elsewhere...

    my_loop.start()
    ```
    """

    def decorator(coro: t.Callable[P, t.Awaitable[None]]) -> IntervalLoop[P]:
        return IntervalLoop(coro, seconds=seconds, minutes=minutes, hours=hours, days=days, run_on_start=run_on_start)

    return decorator


def cron_loop(
    cron_format: str, timezone: datetime.timezone = datetime.timezone.utc
) -> t.Callable[[t.Callable[P, t.Awaitable[None]]], CronLoop[P]]:
    """Decorator to create a [`CronLoop`][arc.utils.loops.CronLoop] out of a coroutine function.

    !!! warning
        To use this loop, you must install arc with the `cron` extra.

        ```sh
        pip install hikari-arc[cron]
        ```

    Parameters
    ----------
    cron_format : str
        The cron format to use. See https://en.wikipedia.org/wiki/Cron for more information.
    timezone : datetime.timezone
        The timezone to use for the cron loop. Defaults to UTC.

    Returns
    -------
    Callable[[t.Callable[P, t.Awaitable[None]]], CronLoop[P]]
        The decorator.

    Raises
    ------
    ImportError
        If the `croniter` package is not installed.
    croniter.CroniterBadCronError
        If the cron format is invalid.
    TypeError
        If the decorated function is not a coroutine function.

    Example
    --------
    ```py
    import arc

    # Run every 5th minute of every hour.
    @arc.utils.cron_loop("*/5 * * * *")
    async def my_loop():
        print("Hello, world!")

    # Elsewhere...

    my_loop.start()
    ```
    """

    def decorator(coro: t.Callable[P, t.Awaitable[None]]) -> CronLoop[P]:
        return CronLoop(coro, cron_format, timezone=timezone)

    return decorator


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import abc
import enum
import typing as t

import attr

if t.TYPE_CHECKING:
    import hikari

    from arc.abc.command import CommandProto
    from arc.abc.option import OptionBase
    from arc.context.base import Context

__all__ = (
    "CommandLocaleRequest",
    "CustomLocaleRequest",
    "LocaleRequest",
    "LocaleRequestType",
    "LocaleResponse",
    "OptionLocaleRequest",
)


@t.final
class LocaleRequestType(enum.IntEnum):
    """The type of locale request."""

    COMMAND = 0
    """A command locale request."""

    OPTION = 1
    """An option locale request."""

    CUSTOM = 2
    """A custom locale request."""


@attr.define(slots=True)
class LocaleRequest(abc.ABC):
    """The base class for all locale requests."""

    _command: CommandProto = attr.field()
    _locale: hikari.Locale = attr.field()

    @property
    @abc.abstractmethod
    def type(self) -> LocaleRequestType:
        """The type of locale request."""

    @property
    def locale(self) -> hikari.Locale:
        """The locale that is being requested."""
        return self._locale

    @property
    def command(self) -> CommandProto:
        """The command that is requesting localization."""
        return self._command

    @property
    def qualified_name(self) -> t.Sequence[str]:
        """The qualified name of the command that is requesting localization."""
        return self._command.qualified_name


@t.final
@attr.define(slots=True)
class LocaleResponse:
    """The response to a command or option locale request."""

    name: str | None = attr.field(default=None)
    """The localized name of the command or option."""

    description: str | None = attr.field(default=None)
    """The localized description of the command or option."""


@t.final
@attr.define(slots=True)
class CommandLocaleRequest(LocaleRequest):
    """A request to localize a command."""

    _name: str = attr.field()
    _description: str | None = attr.field(default=None)

    @property
    def type(self) -> LocaleRequestType:
        """The type of locale request."""
        return LocaleRequestType.COMMAND

    @property
    def name(self) -> str:
        """The name of the command to be localized."""
        return self._name

    @property
    def description(self) -> str | None:
        """The description of the command to be localized, if any."""
        return self._description


@t.final
@attr.define(slots=True)
class OptionLocaleRequest(LocaleRequest):
    """A request to localize a command option."""

    _name: str = attr.field()
    _description: str = attr.field()
    _option: OptionBase[t.Any] = attr.field()

    @property
    def type(self) -> LocaleRequestType:
        """The type of locale request."""
        return LocaleRequestType.OPTION

    @property
    def option(self) -> OptionBase[t.Any]:
        """The option that is requesting localization."""
        return self._option

    @property
    def name(self) -> str:
        """The name of the option to be localized."""
        return self._name

    @property
    def description(self) -> str:
        """The description of the option to be localized."""
        return self._description


@t.final
@attr.define(slots=True)
class CustomLocaleRequest(LocaleRequest):
    """A custom locale request made by the user."""

    _context: Context[t.Any] = attr.field()
    _key: str = attr.field()

    @property
    def type(self) -> LocaleRequestType:
        """The type of locale request."""
        return LocaleRequestType.CUSTOM

    @property
    def context(self) -> Context[t.Any]:
        """The context that is requesting localization."""
        return self._context

    @property
    def key(self) -> str:
        """The key that is to be localized."""
        return self._key


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

bot = hikari.GatewayBot("...", banner=None)
client = arc.GatewayClient(
    bot,
    autodefer=arc.AutodeferMode.EPHEMERAL,
    default_permissions=hikari.Permissions.MANAGE_GUILD,
    invocation_contexts=[hikari.ApplicationContextType.GUILD],
    integration_types=[hikari.ApplicationIntegrationType.GUILD_INSTALL],
)

plugin = arc.GatewayPlugin(
    "foo",
    default_permissions=hikari.Permissions.NONE,
    integration_types=[hikari.ApplicationIntegrationType.GUILD_INSTALL, hikari.ApplicationIntegrationType.USER_INSTALL],
    invocation_contexts=[
        hikari.ApplicationContextType.BOT_DM,
        hikari.ApplicationContextType.GUILD,
        hikari.ApplicationContextType.PRIVATE_CHANNEL,
    ],
)


@plugin.include
@arc.slash_command("foo", invocation_contexts=[hikari.ApplicationContextType.GUILD])
async def foo(ctx: arc.GatewayContext) -> None:
    await ctx.respond("foo")


@plugin.include
@arc.user_command("bar")
async def bar(ctx: arc.GatewayContext, user: arc.Option[hikari.User, arc.UserParams()]) -> None:
    await ctx.respond(f"bar {user.mention}")


group = plugin.include_slash_group("group", "group description", autodefer=False)


@group.include
@arc.slash_subcommand("qux")
async def qux(ctx: arc.GatewayContext) -> None:
    await ctx.respond("foo")


client.add_plugin(plugin)


@client.include
@arc.message_command("baz", autodefer=True)
async def baz(ctx: arc.GatewayContext, message: hikari.Message) -> None:
    await ctx.respond(f"baz {message.author.mention}")


def test_settings_inheritance() -> None:
    assert foo.default_permissions == hikari.Permissions.NONE
    assert list(foo.integration_types) == [
        hikari.ApplicationIntegrationType.GUILD_INSTALL,
        hikari.ApplicationIntegrationType.USER_INSTALL,
    ]
    assert list(foo.invocation_contexts) == [hikari.ApplicationContextType.GUILD]
    assert foo.autodefer is arc.AutodeferMode.EPHEMERAL

    assert list(bar.integration_types) == [
        hikari.ApplicationIntegrationType.GUILD_INSTALL,
        hikari.ApplicationIntegrationType.USER_INSTALL,
    ]
    assert list(bar.invocation_contexts) == [
        hikari.ApplicationContextType.BOT_DM,
        hikari.ApplicationContextType.GUILD,
        hikari.ApplicationContextType.PRIVATE_CHANNEL,
    ]
    assert bar.default_permissions == hikari.Permissions.NONE
    assert bar.autodefer is arc.AutodeferMode.EPHEMERAL

    assert baz.default_permissions == hikari.Permissions.MANAGE_GUILD
    assert list(baz.integration_types) == [hikari.ApplicationIntegrationType.GUILD_INSTALL]
    assert list(baz.invocation_contexts) == [hikari.ApplicationContextType.GUILD]
    assert baz.autodefer is arc.AutodeferMode.ON

    assert qux.autodefer is arc.AutodeferMode.OFF



import datetime
import typing as t

import hikari
import pytest
from hikari.users import UserImpl
from mock_client import MockClient, MockContext, MockPlugin

import arc

client = MockClient(hikari.GatewayBot(token="amongus", banner=None))


@pytest.fixture
def app() -> hikari.GatewayBot:
    return build_app()


def build_app() -> hikari.GatewayBot:
    return hikari.GatewayBot(token="amongus")


def build_user(app: hikari.GatewayBot, id: hikari.Snowflakeish = 123456789) -> hikari.User:
    return UserImpl(
        app=app,
        id=hikari.Snowflake(id),
        avatar_hash=None,
        banner_hash=None,
        global_name="Padoru",
        accent_color=None,
        flags=hikari.UserFlag.NONE,
        avatar_decoration=None,
        discriminator="1234",
        is_bot=False,
        is_system=False,
        username="Padoru",
    )


def build_member(
    app: hikari.GatewayBot,
    id: hikari.Snowflakeish = 123456789,
    *,
    role_ids: list[hikari.Snowflake] | None = None,
    permissions: hikari.Permissions | None = None,
) -> hikari.InteractionMember:
    return hikari.InteractionMember(
        guild_id=hikari.Snowflake(123456789),
        is_deaf=hikari.UNDEFINED,
        is_mute=hikari.UNDEFINED,
        is_pending=False,
        nickname=None,
        raw_communication_disabled_until=None,
        role_ids=role_ids or [],
        guild_avatar_decoration=None,
        guild_banner_hash=None,
        guild_avatar_hash=None,
        joined_at=datetime.datetime(2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
        permissions=permissions or hikari.Permissions.NONE,
        premium_since=None,
        user=build_user(app, id),
        guild_flags=hikari.GuildMemberFlags.NONE,
    )


def build_resolved(
    members: dict[hikari.Snowflake, hikari.InteractionMember] | None = None,
    roles: dict[hikari.Snowflake, hikari.Role] | None = None,
    channels: dict[hikari.Snowflake, hikari.InteractionChannel] | None = None,
    users: dict[hikari.Snowflake, hikari.User] | None = None,
    attachments: dict[hikari.Snowflake, hikari.Attachment] | None = None,
    messages: dict[hikari.Snowflake, hikari.Message] | None = None,
) -> hikari.ResolvedOptionData:
    return hikari.ResolvedOptionData(
        members=members or {},
        roles=roles or {},
        channels=channels or {},
        users=users or {},
        attachments=attachments or {},
        messages=messages or {},
    )


def build_inter(
    app: hikari.GatewayBot,
    *,
    cmd_name: str,
    options: t.Sequence[hikari.CommandInteractionOption] | None = None,
    resolved: hikari.ResolvedOptionData | None = None,
    author_id: hikari.Snowflakeish = 123456789,
    author_role_ids: list[hikari.Snowflake] | None = None,
    author_perms: hikari.Permissions | None = None,
) -> hikari.CommandInteraction:
    parts = cmd_name.split(" ")

    match len(parts):
        case 1:
            pass
        case 2:
            options = [
                hikari.CommandInteractionOption(
                    name=parts[1], type=hikari.OptionType.SUB_COMMAND, value=None, options=options
                )
            ]
        case 3:
            options = [
                hikari.CommandInteractionOption(
                    name=parts[1],
                    type=hikari.OptionType.SUB_COMMAND_GROUP,
                    value=None,
                    options=[
                        hikari.CommandInteractionOption(
                            name=parts[2], type=hikari.OptionType.SUB_COMMAND, value=None, options=options
                        )
                    ],
                )
            ]
        case _:
            raise ValueError("Invalid command name")

    return hikari.CommandInteraction(
        app=app,
        id=hikari.Snowflake(123456789),
        application_id=hikari.Snowflake(123456789),
        command_id=hikari.Snowflake(123456789),
        registered_guild_id=hikari.Snowflake(123456789),
        command_name=parts[0],
        command_type=hikari.CommandType.SLASH,
        entitlements=[],
        app_permissions=hikari.Permissions.all_permissions(),
        authorizing_integration_owners={hikari.ApplicationIntegrationType.GUILD_INSTALL: hikari.Snowflake(123456789)},
        context=hikari.ApplicationContextType.GUILD,
        options=options or [],
        resolved=resolved,
        type=hikari.InteractionType.APPLICATION_COMMAND,
        token="padoru padoru",
        version=1,
        channel=hikari.InteractionChannel(
            app=app,
            id=hikari.Snowflake(123456789),
            name="test",
            type=hikari.ChannelType.GUILD_TEXT,
            parent_id=None,
            permissions=hikari.Permissions.all_permissions(),
            thread_metadata=None,
        ),
        guild_id=hikari.Snowflake(123456789),
        guild_locale="en-US",
        user=build_user(app, author_id),
        member=build_member(app, author_id, role_ids=author_role_ids or [], permissions=author_perms),
        locale="en-US",
    )


class PluginCanHandleError(Exception):
    pass


class GroupCanHandleError(Exception):
    pass


class SubgroupCanHandleError(Exception):
    pass


class CommandCanHandleError(Exception):
    pass


@client.include
@arc.slash_command("ping")
async def ping(ctx: MockContext) -> None:
    await ctx.respond("Pong!")


@client.include
@arc.with_hook(arc.owner_only)
@arc.slash_command("admin_ping")
async def ping_with_hook(ctx: MockContext) -> None:
    await ctx.respond("Pong!")


@client.include
@arc.with_hook(arc.owner_only)
@arc.slash_command("admin_ping_err")
async def ping_with_hook_err(ctx: MockContext) -> None:
    await ctx.respond("Pong!")


@ping_with_hook_err.set_error_handler
async def ping_with_hook_err_handler(ctx: MockContext, error: Exception) -> None:
    if isinstance(error, arc.NotOwnerError):
        await ctx.respond("❌ You are not the owner of the bot.")


@client.include
@arc.slash_command("sum")
async def sum(ctx: MockContext, a: arc.Option[int, arc.IntParams()], b: arc.Option[int, arc.IntParams()]) -> None:
    await ctx.respond(f"Sum: {a + b}")


plugin = MockPlugin("test")
group = plugin.include_slash_group("group")
subgroup = group.include_subgroup("subgroup")

client.add_plugin(plugin)


@subgroup.include
@arc.slash_subcommand("nested")
async def nested(ctx: MockContext) -> None:
    match ctx.author.id:
        case 1:
            raise CommandCanHandleError
        case 2:
            raise SubgroupCanHandleError
        case 3:
            raise GroupCanHandleError
        case 4:
            raise PluginCanHandleError
        case 5:
            raise Exception
        case _:
            await ctx.respond("Nested command!")


@nested.set_error_handler
async def nested_error_handler(ctx: MockContext, error: Exception) -> None:
    if isinstance(error, CommandCanHandleError):
        await ctx.respond("Command handled error")
        return
    raise error


@subgroup.set_error_handler
async def subgroup_error_handler(ctx: MockContext, error: Exception) -> None:
    if isinstance(error, SubgroupCanHandleError):
        await ctx.respond("Subgroup handled error")
        return
    raise error


@group.set_error_handler
async def group_error_handler(ctx: MockContext, error: Exception) -> None:
    if isinstance(error, GroupCanHandleError):
        await ctx.respond("Group handled error")
        return
    raise error


@plugin.set_error_handler
async def plugin_error_handler(ctx: MockContext, error: Exception) -> None:
    if isinstance(error, PluginCanHandleError):
        await ctx.respond("Plugin handled error")
        return
    raise error


@client.include
@arc.slash_command("converters")
async def converters(
    ctx: MockContext,
    user: arc.Option[hikari.Member, arc.MemberParams()],
    color: arc.Option[hikari.Color, arc.ColorParams()],
    emoji: arc.Option[hikari.Emoji, arc.EmojiParams()],
) -> None:
    assert isinstance(user, hikari.InteractionMember)
    assert isinstance(color, hikari.Color)
    assert isinstance(emoji, hikari.Emoji)
    await ctx.respond("All is well!")


@pytest.mark.asyncio
async def test_ping(app: hikari.GatewayBot) -> None:
    response = await client.push_inter(build_inter(app, cmd_name="ping"))
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "Pong!"


@pytest.mark.asyncio
async def test_sum(app: hikari.GatewayBot) -> None:
    inter = build_inter(
        app,
        cmd_name="sum",
        options=[
            hikari.CommandInteractionOption(name="a", type=hikari.OptionType.INTEGER, value=3, options=None),
            hikari.CommandInteractionOption(name="b", type=hikari.OptionType.INTEGER, value=2, options=None),
        ],
    )
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "Sum: 5"


@pytest.mark.asyncio
async def test_admin_ping(app: hikari.GatewayBot) -> None:
    inter = build_inter(app, cmd_name="admin_ping")
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "Pong!"


@pytest.mark.asyncio
async def test_admin_ping_unprivileged(app: hikari.GatewayBot) -> None:
    inter = build_inter(app, author_id=987654321, cmd_name="admin_ping")
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    # Unhandled error
    assert response.content == "❌ Something went wrong. Please contact the bot developer."


@pytest.mark.asyncio
async def test_admin_ping_errhandler(app: hikari.GatewayBot) -> None:
    inter = build_inter(app, author_id=987654321, cmd_name="admin_ping_err")
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    # Handled error
    assert response.content == "❌ You are not the owner of the bot."


@pytest.mark.asyncio
async def test_nested_errhandler_ok(app: hikari.GatewayBot) -> None:
    inter = build_inter(app, author_id=0, cmd_name="group subgroup nested")
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "Nested command!"


@pytest.mark.asyncio
async def test_nested_errhandler_cmdhandler(app: hikari.GatewayBot) -> None:
    inter = build_inter(app, author_id=1, cmd_name="group subgroup nested")
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "Command handled error"


@pytest.mark.asyncio
async def test_nested_errhandler_subghandler(app: hikari.GatewayBot) -> None:
    inter = build_inter(app, author_id=2, cmd_name="group subgroup nested")
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "Subgroup handled error"


@pytest.mark.asyncio
async def test_nested_errhandler_grphandler(app: hikari.GatewayBot) -> None:
    inter = build_inter(app, author_id=3, cmd_name="group subgroup nested")
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "Group handled error"


@pytest.mark.asyncio
async def test_nested_errhandler_plghandler(app: hikari.GatewayBot) -> None:
    inter = build_inter(app, author_id=4, cmd_name="group subgroup nested")
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "Plugin handled error"


@pytest.mark.asyncio
async def test_nested_errhandler_unhandled(app: hikari.GatewayBot) -> None:
    inter = build_inter(app, author_id=5, cmd_name="group subgroup nested")
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "❌ Something went wrong. Please contact the bot developer."


@pytest.mark.asyncio
async def test_converters(app: hikari.GatewayBot) -> None:
    inter = build_inter(
        app,
        cmd_name="converters",
        resolved=build_resolved(members={hikari.Snowflake(123456789): build_member(app, 123456789)}),
        options=[
            hikari.CommandInteractionOption(
                name="user", type=hikari.OptionType.USER, value=hikari.Snowflake(123456789), options=None
            ),
            hikari.CommandInteractionOption(
                name="color", type=hikari.OptionType.STRING, value="123 123 123", options=None
            ),
            hikari.CommandInteractionOption(
                name="emoji", type=hikari.OptionType.STRING, value="<:padoru:123456789>", options=None
            ),
        ],
    )
    response = await client.push_inter(inter)
    assert isinstance(response, hikari.impl.InteractionMessageBuilder)
    assert response.content == "All is well!"



import hikari

import arc

# Options allow users to provide additional information to slash commands
# Related documentation guide: https://arc.hypergonial.com/guides/options

bot = hikari.GatewayBot("...")
client = arc.GatewayClient(bot)


@client.include
@arc.slash_command("options", "All the options!")
async def all_the_options(
    ctx: arc.GatewayContext,
    # ints and floats support min and max values
    number: arc.Option[int, arc.IntParams("A whole number.", max=10, min=0)],
    # If you set choices, those will be the only valid values for the option
    floating: arc.Option[float, arc.FloatParams("A float.", choices=[1.0, 1.5, 2.0])],
    # Text options can have a max length
    text: arc.Option[str, arc.StrParams("Some text.", max_length=50)],
    # You can also use some common Discord types as options, these will be resolved by Discord
    user: arc.Option[hikari.User, arc.UserParams("A user.")],
    # To make an option not required, set a default value for it
    boolean: arc.Option[bool, arc.BoolParams("A boolean.")] = True,
    # Channel types are resolved based on the type passed to the option
    # This option for example will only allow guild channels that can have messages sent to them
    channel: arc.Option[hikari.TextableGuildChannel | None, arc.ChannelParams("A channel.")] = None,
) -> None:
    await ctx.respond(
        f"Number: `{number}`"
        f"\nFloating: `{floating}`"
        f"\nText: `{text}`"
        f"\nUser: {user.mention}"
        f"\nBoolean: `{boolean}`"
        f"\nChannel: {channel.mention if channel else '`None`'}"
    )


# Note that this is not all possible option types, for more information, see the documentation:
# https://arc.hypergonial.com/guides/options

# Autocomplete


# Define an autocompletion callback
# This can either return a list of the option type, or a list of hikari.CommandChoice
async def provide_opts(data: arc.AutocompleteData[arc.GatewayClient, str]) -> list[str]:
    if data.focused_value and len(data.focused_value) > 20:
        return ["That", "is", "so", "long!"]
    return ["Short", "is", "better!"]


# Only 'str', 'int' and 'float' support autocompletion


@client.include
@arc.slash_command("autocomplete", "Autocomplete options!")
async def autocomplete_command(
    ctx: arc.GatewayContext,
    # Set the 'autocomplete_with' parameter to the function that will be used to autocomplete the option
    complete_me: arc.Option[str, arc.StrParams("I'll complete you!", autocomplete_with=provide_opts)],
) -> None:
    await ctx.respond(f"You wrote: `{complete_me}`")


# Note that 'choices' and 'autocomplete_with' are mutually exclusive, you can only use one of them!

# It is also worth keeping in mind that as opposed to 'choices', the user CAN input values that are not in the list
# returned by the autocompletion function. You should validate your input.

bot.run()
